# Copyright (c) 2015. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function, division, absolute_import

from .compute_cache import clear_cached_objects
from .ensembl_release import EnsemblRelease
from .genome import Genome
from .genome_source import GenomeSource
from .ensembl_release_source import EnsemblReleaseSource
from .gene import Gene
from .gtf import GTF
from .locus import Locus
from .release_info import check_human_release_number
from .search import find_nearest_locus
from .sequence_data import SequenceData
from .transcript import Transcript

_cache = {}

def cached_release(version, auto_download=True):
    """Cached construction of EnsemblRelease objects. It's desirable to reuse
    the same EnsemblRelease object since each one will store a lot of cached
    annotation data in-memory.
    """
    version = check_human_release_number(version)
    key = version, auto_download
    if key not in _cache:
        ensembl = EnsemblRelease(version, auto_download=auto_download)
        _cache[key] = ensembl
    return _cache[key]

ensembl_grch36 = ensembl54 = cached_release(54)  # last release for GRCh36/hg18
ensembl_grch37 = ensembl75 = cached_release(75)  # last release for GRCh37/hg19

ensembl77 = cached_release(77)
ensembl78 = cached_release(78)
ensembl79 = cached_release(79)
ensembl_grch38 = ensembl79  # most recent for GRCh38



# Copyright (c) 2015. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from os.path import basename, join, exists
from shutil import copy2

class GenomeSource(object):
    """
    Represents a URL or local file path of a GTF or FASTA file.
    """
    def __init__(self, path_or_url, reference_name, arg_name=None):
        self.path_or_url = path_or_url
        self.reference_name = reference_name
        self.arg_name = arg_name

    @property
    def original_filename(self):
        return basename(self.path_or_url)

    @property
    def cached_filename(self):
        """
        As opposed to the remote filename or the original filename
        of a local source, this is the filename after downloading/copying
        to the cache directory.

        Unless overriden, these are the same.
        """
        return self.original_filename

    def cached_path(self, cache):
        """
        The full path to the cached file.
        """
        return join(cache.cache_directory_path,
                    self.cached_filename)

    def install_string_console(self):
        if not self.arg_name:
            raise ValueError("Expected GenomeSource to contain an arg_name: %s"
                             % str(self))

        # The shell script expects dashed arguments
        console_arg_name = self.arg_name.replace("_", "-")

        return "pyensembl install --reference-name %s --%s \"%s\"" % (
            self.reference_name, console_arg_name, self.path_or_url)

    def install_string_python(self):
        if not self.arg_name:
            raise ValueError("Expected GenomeSource to contain an arg_name: %s"
                             % str(self))
        return "Genome(reference_name=\"%s\", %s=\"%s\").install()" % (
            self.reference_name, self.arg_name, self.path_or_url)

    def is_url_format(self):
        return "://" in self.path_or_url

    def copy_to_cache_if_needed(self, cache, force):
        """
        Given a path to a file, copy the file to the cache's directory
        and return the path to the new file within the cache's directory.

        If `force` is True, overwrites an existing cache directory file.
        """
        assert not self.is_url_format(), \
            "Copying should not be called on a URL: %s" % self.path_or_url
        new_path = self.cached_path(cache)
        if not exists(new_path) or force:
            copy2(self.path_or_url, new_path)
        return new_path

    def __str__(self):
        return "GenomeSource(path_or_url=%s, reference_name=%s, arg_name=%s)" % (
            self.path_or_url, self.reference_name, self.arg_name)

    def __repr__(self):
        return str(self)

    def __eq__(self, other):
        return (
            other.__class__ is GenomeSource and
            self.path_or_url == other.path_or_url and
            self.reference_name == other.reference_name and
            self.arg_name == other.arg_name)

    def __hash__(self):
        return hash((self.path_or_url, self.reference_name, self.arg_name))



# Copyright (c) 2015. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Contains the Genome class, with its millions of accessors and wrappers
around an arbitrary genomic database.
"""

from __future__ import print_function, division, absolute_import

from glob import glob
import logging
from os.path import join
from os import remove

from .common import (
    require_human_transcript_id,
    require_human_protein_id,
    memoize
)
from .compute_cache import cached_object
from .database import Database
from .exon import Exon
from .gene import Gene
from .gtf import GTF
from .sequence_data import SequenceData
from .transcript import Transcript
from .genome_source import GenomeSource

class Genome(object):
    """
    Bundles together the genomic annotation and sequence data associated with
    a particular genomic database source (e.g. a single Ensembl release) and
    provides a wide variety of helper methods for accessing this data.
    """
    def __init__(self,
                 reference_name,
                 gtf_path_or_url=None,
                 transcript_fasta_path_or_url=None,
                 protein_fasta_path_or_url=None,
                 name=None,
                 version=None,
                 only_human=False,
                 auto_download=False,
                 require_ensembl_ids=True):
        self.reference_name = reference_name
        self.gtf_path_or_url = gtf_path_or_url
        self.transcript_fasta_path_or_url = transcript_fasta_path_or_url
        self.protein_fasta_path_or_url = protein_fasta_path_or_url
        self.name = name
        self.version = version
        self.only_human = only_human
        self.auto_download = auto_download
        self.require_ensembl_ids = require_ensembl_ids

        # GTF object wraps the source GTF file from which we get
        # genome annotations. Presents access to each feature
        # annotations as a pandas.DataFrame.
        self.gtf = self.build_gtf() if self.gtf_path_or_url else None

        # Database object turns the GTF dataframes into sqlite3 tables
        # and wraps them with methods like `query_one`
        self.db = Database(gtf=self.gtf, auto_download=auto_download)

        # get the path for the cDNA and pep FASTA files containing
        # this genome's transcript and protein sequences
        self.transcript_sequences = (self.build_transcript_sequences()
                                     if self.transcript_fasta_path_or_url
                                     else None)
        self.protein_sequences = (self.build_protein_sequences()
                                  if self.protein_fasta_path_or_url
                                  else None)

        self.logger = logging.getLogger()
        self.logger.setLevel(logging.INFO)

    def build_gtf(self):
        return GTF(gtf_source=GenomeSource(path_or_url=self.gtf_path_or_url,
                                           reference_name=self.reference_name,
                                           arg_name="gtf_path_or_url"),
                   auto_download=self.auto_download)

    def build_transcript_sequences(self):
        transcript_fasta_source = GenomeSource(
            path_or_url=self.transcript_fasta_path_or_url,
            reference_name=self.reference_name,
            arg_name="transcript_fasta_path_or_url")
        return SequenceData(
            fasta_source=transcript_fasta_source,
            require_ensembl_ids=self.require_ensembl_ids,
            auto_download=self.auto_download)

    def build_protein_sequences(self):
        protein_fasta_source = GenomeSource(
            path_or_url=self.protein_fasta_path_or_url,
            reference_name=self.reference_name,
            arg_name="protein_fasta_path_or_url")
        return SequenceData(
            fasta_source=protein_fasta_source,
            require_ensembl_ids=self.require_ensembl_ids,
            auto_download=self.auto_download)

    def __str__(self):
        return ("Genome(reference_name=%s, "
                "gtf_path_or_url=%s, "
                "transcript_fasta_path_or_url=%s, "
                "protein_fasta_path_or_url=%s, "
                "name=%s, version=%s, only_human=%s)" % (
                    self.reference_name,
                    self.gtf_path_or_url,
                    self.transcript_fasta_path_or_url,
                    self.protein_fasta_path_or_url,
                    self.name,
                    self.version,
                    self.only_human))

    def __repr__(self):
        return str(self)

    def __eq__(self, other):
        return (
            other.__class__ is Genome and
            self.reference_name == other.reference_name and
            self.gtf_path_or_url == other.gtf_path_or_url and
            self.transcript_fasta_path_or_url == other.transcript_fasta_path_or_url and
            self.protein_fasta_path_or_url == other.protein_fasta_path_or_url and
            self.name == other.name and
            self.version == other.version and
            self.only_human == other.only_human)

    def __hash__(self):
        return hash((self.reference_name, self.gtf_path_or_url,
                     self.transcript_fasta_path_or_url,
                     self.protein_fasta_path_or_url,
                     self.name, self.version, self.only_human))

    def _delete_cached_files(self):
        """
        Any files which start with the same name as our GTF file
        is assumed to be some view of the this genome database's data
        and thus safe to delete.
        """
        if self.gtf:
            base = self.gtf.base_filename()
            dirpath = self.gtf.local_gtf_path()
            for path in glob(join(dirpath, base + "*")):
                logging.info("Deleting cached file %s", path)
                remove(path)

    def clear_cache(self):
        for maybe_fn in self.__dict__.values():
            # clear cache associated with all memoization decorators,
            # GTF and SequenceData objects
            if hasattr(maybe_fn, "clear_cache"):
                maybe_fn.clear_cache()

        self._delete_cached_files()

    def all_feature_values(
            self,
            column,
            feature,
            distinct=True,
            contig=None,
            strand=None):
        """
        Cached lookup of all values for a particular feature property from
        the database, caches repeated queries in memory and
        stores them as a CSV.

        Parameters
        ----------

        column : str
            Name of property (e.g. exon_id)

        feature : str
            Type of entry (e.g. exon)

        distinct : bool, optional
            Keep only unique values

        contig : str, optional
            Restrict query to particular contig

        strand : str, optional
            Restrict results to "+" or "-" strands

        Returns a list constructed from query results.
        """
        if not self.gtf:
            raise ValueError("No GTF supplied to this Genome: %s" %
                             str(self))

        # since we're constructing a list, rather than a DataFrame,
        # we're going to store it using pickling rather than the Pandas
        # CSV serializer. Change the default extension from ".csv" to ".pickle"
        pickle_path = self.gtf.cached_data_file_path(
            feature=feature,
            column=column,
            contig=contig,
            strand=strand,
            distinct=distinct,
            extension=".pickle")

        def run_query():
            results = self.db.query_feature_values(
                column=column,
                feature=feature,
                distinct=distinct,
                contig=contig,
                strand=strand)
            assert isinstance(results, list), \
                "Expected list from Database.query_feature_values, got %s" % (
                    type(results))
            return results

        return cached_object(pickle_path, compute_fn=run_query)

    def install(self):
        """
        Explicitely download and index any data for this genome source that
        is not yet downloaded and/or indexed.
        """
        self.download(force=False)
        self.index(force=False)

    def index(self, force=True):
        """
        Create databases and indices for all data for this genome source

        Parameters
        ----------
        force : bool
            Recreate databases even if already indexed (default = True)

        Raises an error if data is not downloaded.
        """
        if self.gtf:
            self.db.create(force=force)
        if self.transcript_sequences:
            self.transcript_sequences.index(force=force)
        if self.protein_sequences:
            self.protein_sequences.index(force=force)

    def transcript_sequence(self, transcript_id):
        """Return cDNA nucleotide sequence of transcript, or None if
        transcript doesn't have cDNA sequence.
        """
        if not self.transcript_sequences:
            raise ValueError("No transcript FASTA supplied to this Genome: %s" %
                             str(self))

        if self.only_human:
            require_human_transcript_id(transcript_id)
        return self.transcript_sequences.get(transcript_id)

    def protein_sequence(self, protein_id):
        """Return cDNA nucleotide sequence of transcript, or None if
        transcript doesn't have cDNA sequence.
        """
        if not self.protein_sequences:
            raise ValueError("No protein FASTA supplied to this Genome: %s" %
                             str(self))

        if self.only_human:
            require_human_protein_id(protein_id)
        return self.protein_sequences.get(protein_id)

    def download(self, force=True):
        """
        Download all data for this genome source.

        Parameters
        ----------
        force : bool
            Download data even if we already have a local copy.
        """
        if self.gtf:
            self.gtf.download(force=force)
        if self.transcript_sequences:
            self.transcript_sequences.download(force=force)
        if self.protein_sequences:
            self.protein_sequences.download(force=force)

    def genes_at_locus(self, contig, position, end=None, strand=None):
        gene_ids = self.gene_ids_at_locus(
            contig, position, end=end, strand=strand)
        return [self.gene_by_id(gene_id) for gene_id in gene_ids]

    def transcripts_at_locus(self, contig, position, end=None, strand=None):
        transcript_ids = self.transcript_ids_at_locus(
            contig, position, end=end, strand=strand)
        return [
            self.transcript_by_id(transcript_id)
            for transcript_id in transcript_ids
        ]

    def exons_at_locus(self, contig, position, end=None, strand=None):
        exon_ids = self.exon_ids_at_locus(
            contig, position, end=end, strand=strand)
        return [
            self.exon_by_id(exon_id)
            for exon_id in exon_ids
        ]

    def gene_ids_at_locus(self, contig, position, end=None, strand=None):
        return self.db.distinct_column_values_at_locus(
            column="gene_id",
            feature="gene",
            contig=contig,
            position=position,
            end=end,
            strand=strand)

    def gene_names_at_locus(self, contig, position, end=None, strand=None):
        return self.db.distinct_column_values_at_locus(
             column="gene_name",
             feature="gene",
             contig=contig,
             position=position,
             end=end,
             strand=strand)

    def exon_ids_at_locus(self, contig, position, end=None, strand=None):
        return self.db.distinct_column_values_at_locus(
            column="exon_id",
            feature="exon",
            contig=contig,
            position=position,
            end=end,
            strand=strand)

    def transcript_ids_at_locus(self, contig, position, end=None, strand=None):
        return self.db.distinct_column_values_at_locus(
            column="transcript_id",
            feature="transcript",
            contig=contig,
            position=position,
            end=end,
            strand=strand)

    def transcript_names_at_locus(
            self, contig, position, end=None, strand=None):
        return self.db.distinct_column_values_at_locus(
            column="transcript_name",
            feature="transcript",
            contig=contig,
            position=position,
            end=end,
            strand=strand)

    def protein_ids_at_locus(self, contig, position, end=None, strand=None):
        return self.db.distinct_column_values_at_locus(
            column="protein_id",
            feature="transcript",
            contig=contig,
            position=position,
            end=end,
            strand=strand)

    ###################################################
    #
    #         Methods which return Locus objects
    #         containing (contig, start, stop, strand)
    #         of various genomic entities
    #
    ###################################################

    @memoize
    def locus_of_gene_id(self, gene_id):
        """
        Given a gene ID returns Locus with: chromosome, start, stop, strand
        """
        return self.db.query_locus(
            filter_column="gene_id",
            filter_value=gene_id,
            feature="gene")

    @memoize
    def loci_of_gene_names(self, gene_name):
        """
        Given a gene name returns list of Locus objects with fields:
            chromosome, start, stop, strand
        You can get multiple results since a gene might have multiple copies
        in the genome.
        """
        return self.db.query_loci("gene_name", gene_name, "gene")

    @memoize
    def locus_of_transcript_id(self, transcript_id):
        return self.db.query_locus(
            filter_column="transcript_id",
            filter_value=transcript_id,
            feature="transcript")

    @memoize
    def locus_of_exon_id(self, exon_id):
        """
        Given an exon ID returns Locus
        """
        return self.db.query_locus("exon_id", exon_id, feature="exon")

    ###################################################
    #
    #             Gene Info Objects
    #
    ###################################################

    @memoize
    def genes(self, contig=None, strand=None):
        """
        Returns all Gene objects in the database. Can be restricted to a
        particular contig/chromosome and strand by the following arguments:

        Parameters
        ----------
        contig : str
            Only return genes on the given contig.

        strand : str
            Only return genes on this strand.
        """
        gene_ids = self.gene_ids(contig=contig, strand=strand)
        return [self.gene_by_id(gene_id) for gene_id in gene_ids]

    @memoize
    def gene_by_id(self, gene_id):
        """
        Construct a Gene object for the given gene ID.
        """
        field_names = [
            "seqname",
            "start",
            "end",
            "strand",
        ]
        optional_field_names = [
            "gene_name",
            "gene_biotype",
        ]
        # Do not look for gene_name and gene_biotype if they are
        # not in the database.
        field_names.extend([name for name in optional_field_names
                            if self.db.column_exists("gene", name)])
        result = self.db.query_one(
            field_names,
            filter_column="gene_id",
            filter_value=gene_id,
            feature="gene")
        if not result:
            raise ValueError("Gene not found: %s" % (gene_id,))

        gene_name, gene_biotype = None, None
        assert len(result) >= 4 and len(result) <= 6, \
            "Result is not the expected length: %d" % len(result)
        contig, start, end, strand = result[:4]
        if len(result) == 5:
            if "gene_name" in field_names:
                gene_name = result[4]
            else:
                gene_biotype = result[4]
        elif len(result) == 6:
            gene_name, gene_biotype = result[4:]

        return Gene(
            gene_id=gene_id,
            gene_name=gene_name,
            contig=contig,
            start=start,
            end=end,
            strand=strand,
            biotype=gene_biotype,
            genome=self,
            require_valid_biotype=("gene_biotype" in field_names))

    @memoize
    def genes_by_name(self, gene_name):
        """
        Get all the unqiue genes with the given name (there might be multiple
        due to copies in the genome), return a list containing a Gene object
        for each distinct ID.
        """
        gene_ids = self.gene_ids_of_gene_name(gene_name)
        return [self.gene_by_id(gene_id) for gene_id in gene_ids]

    @memoize
    def gene_by_protein_id(self, protein_id):
        """
        Get the gene ID associated with the given protein ID,
        return its Gene object
        """
        gene_id = self.gene_id_of_protein_id(protein_id)
        return self.gene_by_id(gene_id)

    ###################################################
    #
    #             Gene Names
    #
    ###################################################

    @memoize
    def _query_gene_name(self, property_name, property_value, feature_type):
        results = self.db.query(
            select_column_names=["gene_name"],
            filter_column=property_name,
            filter_value=property_value,
            feature=feature_type,
            distinct=True,
            required=True)
        return str(results[0][0])

    @memoize
    def gene_names(self, contig=None, strand=None):
        """
        Return all genes in the database,
        optionally restrict to a chromosome and/or strand.
        """
        return self.all_feature_values(
            column="gene_name",
            feature="gene",
            contig=contig,
            strand=strand)

    @memoize
    def gene_name_of_gene_id(self, gene_id):
        return self._query_gene_name("gene_id", gene_id, "gene")

    @memoize
    def gene_name_of_transcript_id(self, transcript_id):
        return self._query_gene_name(
            "transcript_id", transcript_id, "transcript")

    @memoize
    def gene_name_of_transcript_name(self, transcript_name):
        return self._query_gene_name(
            "transcript_name", transcript_name, "transcript")

    @memoize
    def gene_name_of_exon_id(self, exon_id):
        return self._query_gene_name("exon_id", exon_id, "exon")

    ###################################################
    #
    #             Gene IDs
    #
    ###################################################

    @memoize
    def _query_gene_ids(self, property_name, value, feature="gene"):
        results = self.db.query(
            select_column_names=["gene_id"],
            filter_column=property_name,
            filter_value=value,
            feature=feature,
            distinct=True,
            required=True)
        return [str(result[0]) for result in results if result[0]]

    @memoize
    def gene_ids(self, contig=None, strand=None):
        """
        What are all the gene IDs
        (optionally restrict to a given chromosome/contig and/or strand)
        """
        return self.all_feature_values(
            column="gene_id",
            feature="gene",
            contig=contig,
            strand=strand)

    @memoize
    def gene_ids_of_gene_name(self, gene_name):
        """
        What are the gene IDs associated with a given gene name?
        (due to copy events, there might be multiple genes per name)
        """
        results = self._query_gene_ids("gene_name", gene_name)
        if len(results) == 0:
            raise ValueError("Gene name not found: %s" % gene_name)
        return results

    @memoize
    def gene_id_of_protein_id(self, protein_id):
        """
        What is the gene ID associated with a given protein ID?
        """
        results = self._query_gene_ids("protein_id", protein_id,
                                       feature="CDS")
        if len(results) == 0:
            raise ValueError("Protein ID not found: %s" % protein_id)
        assert len(results) == 1, \
            ("Should have only one gene ID for a given protein ID, "
             "but found %d: %s" % (len(results), results))
        return results[0]

    ###################################################
    #
    #             Transcript Info Objects
    #
    ###################################################

    @memoize
    def transcripts(self, contig=None, strand=None):
        """
        Construct Transcript object for every transcript entry in
        the database. Optionally restrict to a particular
        chromosome using the `contig` argument.
        """
        transcript_ids = self.transcript_ids(contig=contig, strand=strand)
        return [
            self.transcript_by_id(transcript_id)
            for transcript_id in transcript_ids
        ]

    @memoize
    def transcript_by_id(self, transcript_id):
        """Construct Transcript object with given transcript ID"""

        optional_field_names = [
            "transcript_name",
            "transcript_biotype",
        ]
        field_names = [
            "seqname",
            "start",
            "end",
            "strand",
            "gene_id",
        ]
        # Do not look for transcript_name and transcript_biotype if
        # they are not in the database.
        field_names.extend([name for name in optional_field_names
                            if self.db.column_exists("transcript", name)])
        result = self.db.query_one(
            select_column_names=field_names,
            filter_column="transcript_id",
            filter_value=transcript_id,
            feature="transcript",
            distinct=True)
        if not result:
            raise ValueError("Transcript not found: %s" % (transcript_id,))

        transcript_name, transcript_biotype = None, None
        assert len(result) >= 5 and len(result) <= 7, \
            "Result is not the expected length: %d" % len(result)
        contig, start, end, strand, gene_id = result[:5]
        if len(result) == 6:
            if "transcript_name" in field_names:
                transcript_name = result[5]
            else:
                transcript_biotype = result[5]
        elif len(result) == 7:
            transcript_name, transcript_biotype = result[5:]

        return Transcript(
            transcript_id=transcript_id,
            transcript_name=transcript_name,
            contig=contig,
            start=start,
            end=end,
            strand=strand,
            biotype=transcript_biotype,
            gene_id=gene_id,
            genome=self,
            require_valid_biotype=("transcript_biotype" in field_names))

    @memoize
    def transcripts_by_name(self, transcript_name):
        transcript_ids = self.transcript_ids_of_transcript_name(transcript_name)
        return [
            self.transcript_by_id(transcript_id)
            for transcript_id in transcript_ids
        ]

    @memoize
    def transcript_by_protein_id(self, protein_id):
        transcript_id = self.transcript_id_of_protein_id(protein_id)
        return self.transcript_by_id(transcript_id)

    ###################################################
    #
    #            Transcript Names
    #
    ###################################################

    @memoize
    def _query_transcript_names(self, property_name, value):
        results = self.db.query(
            select_column_names=["transcript_name"],
            filter_column=property_name,
            filter_value=value,
            feature="transcript",
            distinct=True,
            required=True)
        return [result[0] for result in results]

    @memoize
    def transcript_names(self, contig=None, strand=None):
        """
        What are all the transcript names in the database
        (optionally, restrict to a given chromosome and/or strand)
        """
        return self.all_feature_values(
            column="transcript_name",
            feature="transcript",
            contig=contig,
            strand=strand)

    @memoize
    def transcript_names_of_gene_name(self, gene_name):
        return self._query_transcript_names("gene_name", gene_name)

    @memoize
    def transcript_name_of_transcript_id(self, transcript_id):
        transcript_names = self._query_transcript_names(
            "transcript_id", transcript_id)
        if len(transcript_names) == 0:
            raise ValueError(
                "No transcript names for transcript ID = %s" % transcript_id)
        assert len(transcript_names) == 1, \
            "Multiple transcript names for transcript ID = %s" % transcript_id
        return transcript_names[0]

    ###################################################
    #
    #            Transcript IDs
    #
    ###################################################

    @memoize
    def _query_transcript_ids(
            self,
            property_name,
            value,
            feature="transcript"):
        results = self.db.query(
            select_column_names=["transcript_id"],
            filter_column=property_name,
            filter_value=value,
            feature=feature,
            distinct=True,
            required=True)
        return [result[0] for result in results]

    @memoize
    def transcript_ids(self, contig=None, strand=None):
        return self.all_feature_values(
            column="transcript_id",
            feature="transcript",
            contig=contig,
            strand=strand)

    @memoize
    def transcript_ids_of_gene_id(self, gene_id):
        return self._query_transcript_ids("gene_id", gene_id)

    @memoize
    def transcript_ids_of_gene_name(self, gene_name):
        return self._query_transcript_ids("gene_name", gene_name)

    @memoize
    def transcript_ids_of_transcript_name(self, transcript_name):
        return self._query_transcript_ids("transcript_name", transcript_name)

    @memoize
    def transcript_ids_of_exon_id(self, exon_id):
        return self._query_transcript_ids("exon_id", exon_id)

    @memoize
    def transcript_id_of_protein_id(self, protein_id):
        """
        What is the transcript ID associated with a given protein ID?
        """
        results = self._query_transcript_ids("protein_id", protein_id,
                                             feature="CDS")
        if len(results) == 0:
            raise ValueError("Protein ID not found: %s" % protein_id)
        assert len(results) == 1, \
            ("Should have only one transcript ID for a given protein ID, "
             "but found %d: %s" % (len(results), results))
        return results[0]

    ###################################################
    #
    #             Exon Info Objects
    #
    ###################################################

    @memoize
    def exons(self, contig=None, strand=None):
        """
        Create exon object for all exons in the database, optionally
        restrict to a particular chromosome using the `contig` argument.
        """
        # DataFrame with single column called "exon_id"
        exon_ids = self.exon_ids(contig=contig, strand=strand)
        return [
            self.exon_by_id(exon_id)
            for exon_id in exon_ids
        ]

    @memoize
    def exon_by_id(self, exon_id):
        """Construct an Exon object from its ID by looking up the exon"s
        properties in the given Database.
        """
        field_names = [
            "seqname",
            "start",
            "end",
            "strand",
            "gene_name",
            "gene_id",
        ]

        contig, start, end, strand, gene_name, gene_id = self.db.query_one(
            select_column_names=field_names,
            filter_column="exon_id",
            filter_value=exon_id,
            feature="exon",
            distinct=True)

        return Exon(
            exon_id=exon_id,
            contig=contig,
            start=start,
            end=end,
            strand=strand,
            gene_name=gene_name,
            gene_id=gene_id)

    ###################################################
    #
    #                Exon IDs
    #
    ###################################################

    @memoize
    def _query_exon_ids(self, property_name, value):
        results = self.db.query(
            select_column_names=["exon_id"],
            filter_column=property_name,
            filter_value=value,
            feature="exon",
            distinct=True,
            required=True)
        return [result[0] for result in results]

    @memoize
    def exon_ids(self, contig=None, strand=None):
        return self.all_feature_values(
            column="exon_id",
            feature="exon",
            contig=contig,
            strand=strand)

    @memoize
    def exon_ids_of_gene_id(self, gene_id):
        return self._query_exon_ids("gene_id", gene_id)

    @memoize
    def exon_ids_of_gene_name(self, gene_name):
        return self._query_exon_ids("gene_name", gene_name)

    @memoize
    def exon_ids_of_transcript_name(self, transcript_name):
        return self._query_exon_ids("transcript_name", transcript_name)

    @memoize
    def exon_ids_of_transcript_id(self, transcript_id):
        return self._query_exon_ids("transcript_id", transcript_id)



from __future__ import absolute_import

from . import data_path
from pyensembl import Genome
from nose.tools import eq_

UCSC_GENCODE_PATH = data_path("gencode.ucsc.small.gtf")
UCSC_REFSEQ_PATH = data_path("refseq.ucsc.small.gtf")

def test_ucsc_gencode():
    """
    Testing with a small GENCODE GTF file downloaded from 
    http://genome.ucsc.edu/cgi-bin/hgTables
    """
    genome = Genome("GRCh38", gtf_path_or_url=UCSC_GENCODE_PATH)
    genome.install()
    eq_(len(genome.genes()), 7)
    eq_(len(genome.transcripts()), 7)

    gene_uc001aak4 = genome.gene_by_id("uc001aak.4")
    eq_(gene_uc001aak4.id, "uc001aak.4")
    eq_(gene_uc001aak4.name, None)
    eq_(gene_uc001aak4.biotype, None)

    gene_1_17369 = genome.genes_at_locus(1, 17369)
    eq_(gene_1_17369[0].id, "uc031tla.1")

    transcript_1_30564 = genome.transcripts_at_locus(1, 30564)
    eq_(transcript_1_30564[0].id, "uc057aty.1")

def test_ucsc_refseq():
    """
    Testing with a small RefSeq GTF file downloaded from
    http://genome.ucsc.edu/cgi-bin/hgTables
    """
    genome = Genome("GRCh38", gtf_path_or_url=UCSC_REFSEQ_PATH)
    genome.install()
    eq_(len(genome.genes()), 2)
    eq_(len(genome.transcripts()), 2)

    genes_at_locus = genome.genes_at_locus(1, 67092176)
    eq_(len(genes_at_locus), 2)
    ids = set([gene.id for gene in genes_at_locus])
    eq_(set(["NM_001276352", "NR_075077"]), ids)



# Copyright (c) 2015. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .genome_source import GenomeSource

class EnsemblReleaseSource(GenomeSource):
    """
    Represents a URL or local file path of a GTF or FASTA file
    within an Ensembl release. Also handles some Ensembl-specific
    local filename transformation and generates Ensembl-specific
    install messages.
    """
    def __init__(self,
                 url,
                 release,
                 file_type,
                 reference_name):
        self.release = release
        assert file_type in ["gtf", "fa"], \
            "Expected a .gtf or .fa file, but got: %s" % file_type
        self.file_type = file_type
        GenomeSource.__init__(self,
                              path_or_url=url,
                              reference_name=reference_name)

    @property
    def original_filename(self):
        original_filename = super(EnsemblReleaseSource, self).original_filename
        assert original_filename.endswith(".%s.gz" % self.file_type), \
            "Expected remote GTF file %s to end with '.%s.gz'" % (
                original_filename, self.file_type)
        return original_filename

    @property
    def cached_filename(self):
        """
        We sometimes need to add the release number to a cached FASTA filename
        since some Ensembl releases only have the genome name in the FASTA
        filename but still differ subtly between releases.
        For example, a transcript ID may be missing in Ensembl 75 but present
        in 76, though both have the same FASTA filename
        """
        if ".%d." % self.release in self.original_filename:
            return self.original_filename

        filename_parts = self.original_filename.split(".%s." % self.file_type)
        assert len(filename_parts) == 2, \
            "Expected remote filename %s to contain '.%s.gz'" % (
                self.original_filename, self.file_type)
        return "".join([
            filename_parts[0],
            ".%d.fa." % self.release,
            filename_parts[1]])

        GenomeSource.__init__(self,
                              gtf_path=gtf_url,
                              transcript_fasta_path=transcript_fasta_url,
                              protein_fasta_path=protein_fasta_url)

    def install_string_console(self):
        return "pyensembl install --release %d" % self.release

    def install_string_python(self):
        return "EnsemblRelease(%d).install()" % self.release

    def __str__(self):
        return "EnsemblReleaseSource(%s)" % self._arg_list_str()

    def __repr__(self):
        return str(self)

    def __eq__(self, other):
        return (
            other.__class__ is EnsemblReleaseSource and
            self.path_or_url == other.path_or_url and
            self.release == other.release and
            self.file_type == other.file_type and
            self.reference_name == other.reference_name)

    def __hash__(self):
        return hash((self.path_or_url, self.release, self.file_type,
                     self.reference_name))



#!/usr/bin/env python

# Copyright (c) 2015. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Manipulate pyensembl's local cache.

    %(prog)s {install,download,index} [--release XXX ...]

To install the latest Ensembl release:
    %(prog)s install

To install particular Ensembl release(s):
    %(prog)s install --release 75 77

To install any genome:
    %(prog)s install --reference_name "GRCh38" --gtf-path_or_url URL_OR_PATH --transcript-fasta-path_or_url URL_OR_PATH --protein-fasta-path-or-url URL_OR_PATH

"""

from __future__ import absolute_import
import argparse
from .ensembl_release import EnsemblRelease
from .genome import Genome
from .release_info import MAX_ENSEMBL_RELEASE

def run():
    parser = argparse.ArgumentParser(usage=__doc__)
    root_group = parser.add_mutually_exclusive_group()
    release_group = root_group.add_argument_group()
    release_group.add_argument(
        "--release",
        type=int,
        nargs="+",
        default=None,
        help=("Ensembl release. Defaults to latest release %(default)s. "
              "Multiple releases may be specified."))

    path_group = root_group.add_argument_group()
    path_group.add_argument(
        "--reference-name",
        type=str,
        default=None,
        help="Name of the reference, e.g. GRCh38") 
    path_group.add_argument(
        "--gtf-path-or-url",
        type=str,
        default=None,
        help="URL or local path to a GTF file containing annotations.")
    path_group.add_argument(
        "--transcript-fasta-path-or-url",
        type=str,
        default=None,
        help="URL or local path to a FASTA file containing transcript data.")
    path_group.add_argument(
        "--protein-fasta-path-or-url",
        type=str,
        default=None,
        help="URL or local path to a FASTA file containing protein data.")

    parser.add_argument("action", choices=("install", "download", "index"),
        help="\"install\" will download and index any data that is  not "
        "currently downloaded or indexed. \"download\" will download data, "
        "regardless of whether it is already downloaded. \"index\" will index "
        "data, regardless of whether it is already indexed, and will raise "
        "an error if the data is not already downloaded.")

    args = parser.parse_args()

    genomes = []
    # If specific genome source URLs are provided, use those
    if args.gtf_path_or_url or args.transcript_fasta_path_or_url or args.protein_fasta_path_or_url:
        assert not args.release, ("A release cannot be specified if "
                                  "specific paths are specified.")
        assert args.reference_name, "Must specify a reference name"
        genomes.append(Genome(
            args.reference_name,
            gtf_path_or_url=args.gtf_path_or_url,
            transcript_fasta_path_or_url=args.transcript_fasta_path_or_url,
            protein_fasta_path_or_url=args.protein_fasta_path_or_url))
    # Otherwise, use Ensembl release information
    else:
        if not args.release:
            genomes.append(EnsemblRelease(MAX_ENSEMBL_RELEASE))
        else:
            for release in args.release:
                genomes.append(EnsemblRelease(release))

    for genome in genomes:
        assert args.action in ["install", "download", "index"], \
            "Invalid action: %s" % args.action
        if args.action == "install":
            genome.install()
        elif args.action == "download":
            genome.download()
        else:
            genome.index()


if __name__ == "__main__":
    run()



# Copyright (c) 2015. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Contains the EnsemblRelease class, which extends the Genome class
to be specific to (a particular release of) Ensembl.
"""

from .genome import Genome
from .gtf import GTF
from .ensembl_release_source import EnsemblReleaseSource
from .release_info import (
    check_release_number,
    MAX_ENSEMBL_RELEASE,
    which_human_reference_name
)
from .sequence_data import SequenceData
from .url_templates import ENSEMBL_FTP_SERVER, make_gtf_url, make_fasta_url


class EnsemblRelease(Genome):
    """
    Bundles together the genomic annotation and sequence data associated with
    a particular release of the Ensembl database.
    """
    def __init__(self,
                 release=MAX_ENSEMBL_RELEASE,
                 species="homo_sapiens",
                 server=ENSEMBL_FTP_SERVER,
                 reference_name=None,
                 auto_download=False):
        self.release = check_release_number(release)
        self.species = species
        self.server = server

        self.gtf_url = make_gtf_url(
            ensembl_release=release,
            species=species,
            server=server)
        self.transcript_fasta_url = make_fasta_url(
            ensembl_release=release,
            species=species,
            sequence_type="cdna",
            server=server)
        self.protein_fasta_url = make_fasta_url(
            ensembl_release=release,
            species=species,
            sequence_type="pep",
            server=server)
        only_human = species == "homo_sapiens"

        if not reference_name:
            if only_human:
                reference_name = which_human_reference_name(self.release)
            else:
                raise ValueError("Must provide a reference_name for "
                                 "non-human Ensembl releases.")
        self.reference_name = reference_name

        Genome.__init__(self,
                        reference_name=self.reference_name,
                        gtf_path_or_url=self.gtf_url,
                        transcript_fasta_path_or_url=self.transcript_fasta_url,
                        protein_fasta_path_or_url=self.protein_fasta_url,
                        name="Ensembl",
                        version=self.release,
                        only_human=only_human,
                        auto_download=auto_download,
                        require_ensembl_ids=True)

    def build_gtf(self):
        return GTF(gtf_source=EnsemblReleaseSource(url=self.gtf_url,
                                                   release=self.release,
                                                   file_type="gtf",
                                                   reference_name=self.reference_name),
                   auto_download=self.auto_download)

    def build_transcript_sequences(self):
        transcript_fasta_source = EnsemblReleaseSource(
            url=self.transcript_fasta_url,
            release=self.release,
            file_type="fa",
            reference_name=self.reference_name)
        return SequenceData(
            fasta_source=transcript_fasta_source,
            require_ensembl_ids=self.require_ensembl_ids,
            auto_download=self.auto_download)

    def build_protein_sequences(self):
        protein_fasta_source = EnsemblReleaseSource(
            url=self.protein_fasta_url,
            release=self.release,
            file_type="fa",
            reference_name=self.reference_name)
        return SequenceData(
            fasta_source=protein_fasta_source,
            require_ensembl_ids=self.require_ensembl_ids,
            auto_download=self.auto_download)

    def __str__(self):
        return "EnsemblRelease(release=%d, species=%s)" % (
            self.release,
            self.species)

    def __repr__(self):
        return str(self)

    def __eq__(self, other):
        return (
            other.__class__ is EnsemblRelease and
            self.release == other.release and
            self.species == other.species)

    def __hash__(self):
        return hash((self.release, self.species))
